# -*- coding: utf-8 -*-

"""
This module provide steganography functionality, using picture file as medium
Steganography format:
+ Bcrypt hash of password: bytes, 60 bytes.
+ Salt for encryption key: 16 bytes (AES case).
+ Data size: integer, 64 bytes (zero-padded as needed).
+ Data: bytes, as much as needed.
"""

import argparse

import steganosaurus.constants as constants
import steganosaurus.hash_utils as hash_utils
from cryptography.fernet import InvalidToken
from steganosaurus.fernet_cryptor import FernetCryptor
from steganosaurus.otp_cryptor import OTPCryptor
from steganosaurus.picture_modifier import PictureModifier


class Steganography:
    '''
    Using steganography to hide secret data into picture.
    Data can be encrypted using Fernet encryption or OTP encryption.

    This package can be included into other project or called straight from command line.

    Supported encryption: 
    + Fernet encryption (AES-128) 
    + OTP encryption.

    Supported output format:
    + BMP and PNG image.

    Different between ste_password and password:
    + ste_password is used to check if medium contains hidden data.
    + password is encryption password.
    + For Fernet encryption, ste_password == password.
    + For OTP encryption, ste_password is SHA1 hash of pad, password is path to pad.
    '''

    def __init__(self, medium_modifer, encryptor):
        '''
        Initialization method,
        set medium_modifer object and encryptor here.
        '''

        self.medium_modifer = medium_modifer
        self.encryptor = encryptor

    def hide_file(self, password, medium, secret, format):
        '''
        Hide secret file into medium.
        Return binary value of medium file after adding secret file.
        '''

        self.medium_modifer.load_medium(medium)
        ste_password = self.encryptor.get_steganography_password(password)
        ste_password_hash = hash_utils.calculate_hash(ste_password, constants.HASH_FACTOR)
        self.medium_modifer.hide_data(ste_password_hash)

        # Only Fernet encryption needs salt for pbkdf, OTP does not need salt.
        if self.encryptor.need_pbkdf_salt:
            pbkdf_salt = self.encryptor.set_pbkdf_salt()
            self.medium_modifer.hide_data(pbkdf_salt)

        encrypted = self.encryptor.encrypt(password, secret)
        encrypted_size = len(encrypted).to_bytes(constants.DATA_LENGTH_SIZE, byteorder='big')
        self.medium_modifer.hide_data(encrypted_size)
        self.medium_modifer.hide_data(encrypted)
        
        return self.medium_modifer.export_medium(format)

    def get_file(self, password, medium):
        '''
        Retrieve secret file from medium.
        Return binary value of secret file retrieve from medium.
        '''

        self.medium_modifer.load_medium(medium)
        ste_password_hash = self.medium_modifer.get_data(constants.BCRYPT_HASH_SIZE)
        ste_password = self.encryptor.get_steganography_password(password)

        if not hash_utils.verify_hash(ste_password, ste_password_hash):
            raise PermissionError(constants.STEGANOGRAPHY_NOT_FOUND_ERROR)

        if self.encryptor.need_pbkdf_salt:
            pbkdf_salt = self.medium_modifer.get_data(constants.FERNET_SALT_SIZE)
            self.encryptor.pbkdf_salt = pbkdf_salt

        # Read 8 bytes and conver to int, this is the size of encrypted data.
        encrypted_file_size = self.medium_modifer.get_data(constants.DATA_LENGTH_SIZE)
        encrypted_file_size = int.from_bytes(encrypted_file_size, byteorder='big')
        encrypted = self.medium_modifer.get_data(encrypted_file_size)

        return self.encryptor.decrypt(password, encrypted)


def main():
    '''
    Entry point of script when run from command line.
    '''

    parser=argparse.ArgumentParser()
    parser.add_argument('--mode', help='Encryption/Decryption flag. Allowed value: encrypt/decrypt')
    parser.add_argument('--encryption', help='Encryption method. Use \'aes\' for Fernet encryption or use \'otp\' for OTP encryption')
    parser.add_argument('--key_stretch', help='Number of iteration use when deriving Fernet key from password. Must be a number. Required if encryption mode is aes')
    parser.add_argument('--level', help='Number of low-bit in each pixel uses to store secret data. Must be between 1~8')
    parser.add_argument('--medium', help='Path to medium file')
    parser.add_argument('--password', help='Password for Fernet encryption or path to OTP pad')
    parser.add_argument('--secret', help='Path to secret file. Required in encryption mode')
    parser.add_argument('--result', help='Path to result file')
    parser.add_argument('--format', help='Format of picture with steganography. Allowed value: BMP/PNG (case insensitive). Required in encryption mode')
    args=parser.parse_args()

    # Input check.
    if not verify_argument(args):
        print(constants.HELP_MSG)
        return

    try:
        # Fernet encryption use args.password to derive encryption key.
        if args.encryption == constants.MODE_FERNET:
            encryptor = FernetCryptor(int(args.key_stretch))
            password = args.password
        
        # OTP encryption read binary file at args.password and use it as encryption pad.
        else:
            encryptor = OTPCryptor()
            with open(args.password, 'rb') as f:
                password = f.read()

        modifier = PictureModifier(int(args.level))
        steganography = Steganography(modifier, encryptor)
        with open(args.medium, 'rb') as f:
            medium_data = f.read()

        if args.mode == constants.MODE_ENCRYPT:
            with open(args.secret, 'rb') as f:
                secret_data = f.read()
            file_with_secret = steganography.hide_file(password, medium_data, secret_data, args.format)
            with open(args.result, 'wb') as f:
                f.write(file_with_secret) 

        else:
            secret_data = steganography.get_file(password, medium_data)
            with open(args.result, 'wb') as f:
                f.write(secret_data)

    except FileNotFoundError as fe:
        print('Cannot open file, please check if file path is correct')
        print(str(fe))

    except InvalidToken as it:
        print('Decryption password is incorrect.')
        print(str(it))

    except ValueError as ve:
        print('An ValueError occurred. Please check value of all parameters')
        print(str(ve))

    except TypeError as te:
        print('An TypeError occurred. Please check type of all parameters')
        print(str(te))

    except Exception as e:
        print('An unknown error occurred.')
        print(str(e))


def verify_argument(args):
    '''
    Verify that the list of arguments passing to steganography.py is valid.
    '''

    # Verify all required arguments existed.
    if not (args.mode and args.encryption and args.level \
                and args.medium and args.password and args.result):
        return False

    if args.mode not in [constants.MODE_ENCRYPT, constants.MODE_DECRYPT]:
        return False

    if args.encryption not in [constants.MODE_FERNET, constants.MODE_OTP]:
        return False

    # In encryption mode, secret file path and result format is required.
    if args.mode == constants.MODE_ENCRYPT:
        if not (args.secret and args.format) or \
            args.format.lower() not in [constants.FORMAT_BMP, constants.FORMAT_PNG]:
            return False

    # When encryption mode is Fernet (AES-128), key-stretch iteration is required.
    if args.encryption == constants.MODE_FERNET and \
        not is_positive_number(args.key_stretch):
        return False

    if not is_positive_number(args.level):
        return False
    elif int(args.level) < 1 or int(args.level) > 8:
        return False
    
    return True


def is_positive_number(input):
    '''
    Check if an object can be converted into a positive number.
    Better to ask for forgiveness than for permission. 
    '''

    try:
        int_val = int(input)

    except:
        return False

    return int_val > 0


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

'''
Provide hashing function for steganosaurus.
Right now the only supported hashing method is bcrypt.
'''

import base64

import bcrypt
import steganosaurus.constants as constants


def calculate_hash(password, work_factor):
    '''
    Hash password using bcrypt with given work factor.
    '''

    salt = bcrypt.gensalt(work_factor)
    return bcrypt.hashpw(password.encode('utf-8'), salt)


def verify_hash(password, hash):
    '''
    Verify if given password is correct using bcrypt hash.
    '''

    return hash == bcrypt.hashpw(password.encode(), hash)



# -*- coding: utf-8 -*-

"""
This module provides encryption/decryption for steganosaurus using One-Time-Pad.
"""

import hashlib

import steganosaurus.constants as constants


class OTPCryptor:
    '''
    Encrypt/Decrypt data using One-Time-Pad.
    The pad MUST be true random or we won't get the full benefit of OTP.
    '''

    def __init__(self):
        '''
        Initialize function.
        Set has_pbkdf_salt flag to False because OTP does not need PBDKF.
        '''

        self.need_pbkdf_salt = False


    def encrypt(self, key, data):
        '''
        Encrypt data using OTP by xor-ing it with given key (pad).
        Pad and data must be of the same length for OTP to be fully secured.
        '''

        return self.xor_data(key, data)

    def decrypt(self, key, blob):
        '''
        Decrypt data using OTP by xor-ing it with given key (pad).
        Pad and data must be of the same length for OTP to be fully secured.
        '''

        return self.xor_data(key, blob)

    def xor_data(self, key, data):
        '''
        Xor data with given key (pad).
        Pad and data must be of the same length for OTP to be fully secured.
        '''

        if len(key) != len(data):
            raise ValueError(constants.PAD_LENGTH_ERROR)

        return bytes(a ^ b for (a, b) in zip(key, data))

    def get_steganography_password(self, pad):
        '''
        Create a password to verify if medium contains steganography or not.
        Return hex value of SHA-1 hash of pad.
        Because pad is true random, SHA-1 is secure enough.
        '''

        return hashlib.sha1(pad).hexdigest()


# -*- coding: utf-8 -*-

"""
Constants for steganosaurus.
"""

# Steganography format
BCRYPT_HASH_SIZE = 60
PBKDF_SALT_SIZE = 16
DATA_LENGTH_SIZE = 8

# Constants
KEY_STRETCH_ITERATION = 100000
FERNET_KEY_SIZE = 32
FERNET_SALT_SIZE = 16
DEFAULT_STEGANOGRAPHY_LEVEL = 2
MIN_STEGANOGRAPHY_LEVEL = 1
MAX_STEGANOGRAPHY_LEVEL = 8
HASH_FACTOR = 8

# Error messages
SALT_LENGTH_ERROR_MSG = 'Cannot generate salt because length is invalid'
SALT_ERROR_MESSAGE = 'Cannot generate encryption key because salt is invalid'
PASSWORD_ERROR_MESSAGE = 'Cannot generate encryption key because password is invalid'
PAD_LENGTH_ERROR = 'When using OTP, pad and data must be of the same length'
WRONG_KEY_MSG = 'Cannot decrypt data because the key is incorrect'
STEGANOGRAPHY_LEVEL_VALUE_ERROR_MSG = 'Steganography level must be between 1 ~ 8'
STEGANOGRAPHY_LEVEL_TYPE_ERROR_MSG = 'Steganography level must be a number'
PICTURE_NOT_LOADED_ERROR_MSG = 'Please load original picture first'
MAX_COLOR_INDEX = 3
INPUT_DATA_ERROR_MSG = 'Cannot hide data into image, data must be in bytes format'
DATA_SIZE_ERROR_MSG = 'Data is too big, please choose a bigger medium'
GET_DATA_PARA_TYPE_ERROR_MSG = 'Size parameter of get_data must be integer'
GET_DATA_PARA_VALUE_ERROR_MSG = 'Size parameter of get_data is too big'
GET_DATA_PARA_VALUE_NEGATIVE_MSG = 'Size parameter of get_data must be bigger than 0'
STEGANOGRAPHY_NOT_FOUND_ERROR = 'Medium does not contains steganography or password is incorrect'


# Commandline option
MODE_ENCRYPT = 'encrypt'
MODE_DECRYPT = 'decrypt'
MODE_FERNET = 'aes'
MODE_OTP = 'otp'
FORMAT_BMP = 'bmp'
FORMAT_PNG = 'png'

# Help message for steganosaurus.
HELP_MSG = '\n' + \
            'Input parameter was incorrect\n' + \
            'Try \'python steganography.py -h\' for more information'



from setuptools import setup

requires = [
    'cryptography>=2.3.1',
    'bitarray >= 0.8.3',
    'Pillow>=5.3.0',
    'bcrypt>=3.1.4'
]

test_requirements = [
    'pytest-mock',
    'pytest>=3.7.1'
]

with open('README.md', 'r') as fh:
    long_description = fh.read()

setup(
   name='steganosaurus',
   version='1.0.1',
   description='Hide data into image file',
   author='Nguyen Thai Duong',
   author_email='duongnt.bk@gmail.com',
   long_description=long_description,
   long_description_content_type='text/markdown',
   url='https://github.com/duongntbk/steganosaurus',
   packages=['steganosaurus'],  #same as name
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
   install_requires=requires,
   tests_require=test_requirements,
)


# -*- coding: utf-8 -*-

"""
This module provides encryption/decryption for steganosaurus using Fernet encryption.
"""

import base64
import os

from cryptography.fernet import Fernet
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC

import steganosaurus.constants as constants


class FernetCryptor:
    '''
    Encrypt/Decrypt data using Fernet (symmetric encryption).
    Fernet is basically AES128 in CBC mode with a SHA256 HMAC message authentication code.
    Spec: https://github.com/fernet/spec/blob/master/Spec.md
    '''

    def __init__(self, iteration):
        '''
        Initialization method,
        set number of hash iteration for key-stretching function here.
        '''

        self.iteration = iteration
        self.need_pbkdf_salt = True
        self.pbkdf_salt = None

    def generate_salt(self, length):
        '''
        Generate a random bytearray with given length to use as salt.
        '''

        if not isinstance(length, int):
            raise TypeError(constants.SALT_LENGTH_ERROR_MSG)

        if length < 0:
            raise ValueError(constants.SALT_LENGTH_ERROR_MSG)

        return os.urandom(length)

    def generate_key(self, salt, password):
        '''
        Stretch a password in an encryption key using PBKDF2HMAC.
        The number of hash iteration can be set at initialization.
        '''

        if not isinstance(salt, bytes):
            raise TypeError(constants.SALT_ERROR_MESSAGE)

        if len(salt) != constants.FERNET_SALT_SIZE:
            raise ValueError(constants.SALT_ERROR_MESSAGE)

        if not isinstance(password, str):
            raise TypeError(constants.PASSWORD_ERROR_MESSAGE)

        backend = default_backend()
        kdf = PBKDF2HMAC(
            algorithm=hashes.SHA256(),
            length=constants.FERNET_KEY_SIZE,
            salt=salt,
            iterations=self.iteration,
            backend=backend
        )

        return base64.urlsafe_b64encode(kdf.derive(password.encode('utf-8')))

    def encrypt(self, password, data):
        '''
        Encrypt data using Fernet encryption with given key.
        '''

        key = self.generate_key(self.pbkdf_salt, password)
        f = Fernet(key)
        return f.encrypt(data)

    def decrypt(self, password, blob):
        '''
        Decrypt data using Fernet encryption with given key.
        '''

        key = self.generate_key(self.pbkdf_salt, password)
        f = Fernet(key)
        return f.decrypt(blob)

    def get_steganography_password(self, password):
        '''
        Create a password to verify if medium contains steganography or not.
        Because Fernet encryption is key-based, just return the encryption key.
        '''

        return password

    def set_pbkdf_salt(self):
        '''
        Create 16 bytes of random data and set it as salt for key-derived function.
        Return the salt to caller.
        '''

        salt = os.urandom(constants.FERNET_SALT_SIZE)
        self.pbkdf_salt = salt
        return salt


# -*- coding: utf-8 -*-

"""
This module modifies picture to hide data into low-bit position of r, g, b color
"""

import io

import steganosaurus.constants as constants
from bitarray import bitarray
from PIL import Image


class PictureModifier:
    '''
    This class can modify each pixel of picture to hide data.
    '''

    def __init__(self, ste_lvl=constants.DEFAULT_STEGANOGRAPHY_LEVEL):
        '''
        Constructor: 
        + Initialize the number of steganography bit in each byte, max is 8.
        (the higher this value, the more stenography picture will differs from original picture).
        + Initialize object to store original picture.
        + Initialize current read and write offset.
        '''

        if not isinstance(ste_lvl, int):
            raise TypeError(constants.STEGANOGRAPHY_LEVEL_TYPE_ERROR_MSG)

        if ste_lvl < constants.MIN_STEGANOGRAPHY_LEVEL or \
                constants.MAX_STEGANOGRAPHY_LEVEL < ste_lvl:
            raise ValueError(constants.STEGANOGRAPHY_LEVEL_VALUE_ERROR_MSG)

        self.ste_lvl = ste_lvl
        self.curr_write_pixel_index = 0
        self.curr_read_pixel_index = 0
        self.img = None
        self.pixels = None
        self.img_width = 0
        self.img_height = 0

    def load_medium(self, data):
        '''
        Load picture from give path.
        '''

        self.img = Image.open(io.BytesIO(data))
        self.pixels = self.img.load() # Lazy loading.
        self.img_width = self.img.width
        self.img_height = self.img.height

    def export_medium(self, format):
        '''
        Return binary data of medium.
        This method will be called after secret has been hidden into medium.
        '''

        output = io.BytesIO()
        self.img.save(output, format=format)
        return output.getvalue()

    def hide_data(self, data):
        '''
        Hide data into low-bit of pixels.
        '''

        if self.img is None:
            raise ValueError(constants.PICTURE_NOT_LOADED_ERROR_MSG)

        if not isinstance(data, bytes):
            raise TypeError(constants.INPUT_DATA_ERROR_MSG)

        if len(data) * 8 >= self.img_width * self.img_height * constants.MAX_COLOR_INDEX * self.ste_lvl:
            raise ValueError(constants.DATA_SIZE_ERROR_MSG)
        
        bitdata = bitarray()
        bitdata.frombytes(data)
        bit_per_pixel = constants.MAX_COLOR_INDEX * self.ste_lvl
        for i in range(0, len(bitdata), bit_per_pixel):
            self.hide_data_into_pixel(bitdata[i:i+bit_per_pixel]) # No index out-of-range risk because of python magic.

    def hide_data_into_pixel(self, data):
        '''
        Hide data into low-bit of 1 individual pixel.
        '''

        i_height = self.curr_write_pixel_index // self.img_width
        i_width = self.curr_write_pixel_index - self.img_width * i_height
        color = list(self.pixels[i_width, i_height])
        color_index = 0
        depth_index = 0

        for bit in data:
            if depth_index == self.ste_lvl:
                color_index += 1
                depth_index = 0

            # Inlining bit-wise operation to improve performance.
            mask = 1 << depth_index
            color[color_index] &= ~mask
            if bit:
                color[color_index] |= mask
            depth_index += 1

        self.pixels[i_width, i_height] = tuple(color)
        self.curr_write_pixel_index += 1

    def get_data(self, size):
        '''
        Get data from low-bit of pixel.
        Unit of return data and offset is bytes.
        '''

        if self.img is None:
            raise ValueError(constants.PICTURE_NOT_LOADED_ERROR_MSG)

        if not isinstance(size, int):
            raise TypeError(constants.GET_DATA_PARA_TYPE_ERROR_MSG)

        if size <= 0:
            raise ValueError(constants.GET_DATA_PARA_VALUE_NEGATIVE_MSG)

        if self.img_width * self.img_height * constants.MAX_COLOR_INDEX * self.ste_lvl <= size * 8:
            raise ValueError(constants.GET_DATA_PARA_VALUE_ERROR_MSG)

        b_arr = bitarray()
        pixel_bit_capacity = constants.MAX_COLOR_INDEX * self.ste_lvl

        for index in range(0, size * 8, pixel_bit_capacity):
            i_height = self.curr_read_pixel_index // self.img_width
            i_width = self.curr_read_pixel_index - self.img_width * i_height
            pixel = self.pixels[i_width, i_height]
            bit_count = min(pixel_bit_capacity, size * 8 - index)
            pixel_data = self.get_data_from_pixel(pixel, bit_count)
            b_arr.extend(pixel_data)
            self.curr_read_pixel_index += 1

        return b_arr.tobytes()

    def get_data_from_pixel(self, pixel, size):
        '''
        Get data from low-bit of individual pixel.
        '''

        color_index = 0
        depth_index = 0
        b_arr = bitarray()

        for _ in range(size):
            if depth_index == self.ste_lvl:
                color_index += 1
                depth_index = 0

            # Get value of low-bit from pixel.
            # Inlining this process to improve performance.
            mask = 1 << depth_index
            bit_val = (pixel[color_index] & mask) >> depth_index
            # Add retrived bit to array.
            b_arr.append(bit_val)
            depth_index += 1

        return b_arr
